import pytest

from fastaio import Component

pytestmark = pytest.mark.anyio


async def test_component(capsys):

    class Subcomponent0(Component):
        async def start(self):
            self.done()
            print("started0")

        async def stop(self):
            self.done()
            print("stopped0")

    class Subcomponent1(Component):
        async def start(self):
            self.done()
            print("started1")

        async def stop(self):
            self.done()
            print("stopped1")

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())
            self.add_component(Subcomponent1())

    component0 = Component0()

    async with component0:
        pass

    captured = capsys.readouterr()
    assert captured.out in (
        "started0\nstarted1\nstopped0\nstopped1\n",
        "started0\nstarted1\nstopped1\nstopped0\n",
        "started1\nstarted0\nstopped1\nstopped0\n",
        "started1\nstarted0\nstopped0\nstopped1\n",
    )


async def test_add_same_component_name():
    class Subcomponent0(Component):
        pass

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0(), name="subcomponent0")
            self.add_component(Subcomponent0(), name="subcomponent0")

    with pytest.raises(RuntimeError) as excinfo:
        Component0()

    assert str(excinfo.value) == "Component name already exists: subcomponent0"


async def test_component_not_initialized():
    class Component0(Component):
        def __init__(self):
            pass

    component0 = Component0()

    with pytest.raises(RuntimeError) as excinfo:
        async with component0:
            pass  # pragma: no cover

    assert str(excinfo.value) == "You must call super().__init__() in the __init__ method of your component"



import sys

import pytest

from fastaio import Component

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_exception_prepare(capsys):

    class Component0(Component):
        async def prepare(self):
            print("prepare0")
            raise RuntimeError("prepare0")

        async def start(self):
            # should not be called, since prepare failed
            print("start0")  # pragma: no cover
            self.done()  # pragma: no cover

        async def stop(self):
            # should always be called
            print("stop0")
            self.done()

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "prepare0"

    assert capsys.readouterr().out.splitlines() == ["prepare0", "stop0"]


async def test_exception_start(capsys):

    class Component0(Component):
        async def prepare(self):
            print("prepare0")
            self.done()

        async def start(self):
            print("start0")
            raise RuntimeError("start0")

        async def stop(self):
            # should always be called
            print("stop0")
            self.done()

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "start0"

    assert capsys.readouterr().out.splitlines() == ["prepare0", "start0", "stop0"]


async def test_exception_stop(capsys):

    class Component0(Component):
        async def prepare(self):
            print("prepare0")
            self.done()

        async def start(self):
            print("start0")
            self.done()

        async def stop(self):
            print("stop0")
            raise RuntimeError("stop0")

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "stop0"

    assert capsys.readouterr().out.splitlines() == ["prepare0", "start0", "stop0"]


async def test_exception_prepare_stop(capsys):

    class Component0(Component):
        async def prepare(self):
            print("prepare0")
            raise RuntimeError("prepare0")

        async def start(self):
            print("start0")  # pragma: no cover
            self.done()  # pragma: no cover

        async def stop(self):
            print("stop0")
            raise RuntimeError("stop0")

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 2
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "prepare0"
    assert str(excinfo.value.exceptions[1]) == "stop0"

    assert capsys.readouterr().out.splitlines() == ["prepare0", "stop0"]


async def test_exception_start_stop(capsys):

    class Component0(Component):
        async def prepare(self):
            print("prepare0")
            self.done()

        async def start(self):
            print("start0")
            raise RuntimeError("start0")

        async def stop(self):
            print("stop0")
            raise RuntimeError("stop0")

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 2
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "start0"
    assert str(excinfo.value.exceptions[1]) == "stop0"

    assert capsys.readouterr().out.splitlines() == ["prepare0", "start0", "stop0"]


async def test_exception_subcomponent(capsys):

    class Subcomponent0(Component):
        async def prepare(self):
            print("sub prepare0")
            self.done()

        async def start(self):
            print("sub start0")
            raise RuntimeError("sub start0")

        async def stop(self):
            print("sub stop0")
            raise RuntimeError("sub stop0")

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())

        async def prepare(self):
            print("prepare0")
            self.done()

        async def start(self):
            print("start0")
            self.done()

        async def stop(self):
            print("stop0")
            self.done()

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 2
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "sub start0"
    assert str(excinfo.value.exceptions[1]) == "sub stop0"

    assert capsys.readouterr().out.splitlines() == [
        "prepare0",
        "sub prepare0",
        "start0",
        "sub start0",
        "stop0",
        "sub stop0",
    ]



import sys

import pytest

from anyio import TASK_STATUS_IGNORED, create_task_group, sleep
from anyio.abc import TaskStatus
from fastaio import Component

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_task(capsys):

    class Resource0:
        pass

    async def task(message: str, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        print(message)
        task_status.started()
        await sleep(float("inf"))

    class Subcomponent0(Component):
        async def prepare(self):
            async with create_task_group() as self.tg1:
                await self.tg1.start(task, "prepare0", name="prepare0")
                self.done()
                print("prepared0")

        async def start(self):
            async with create_task_group() as self.tg2:
                await self.tg2.start(task, "start0", name="start0")
                self.done()
                print("started0")
                self.resource0 = Resource0()
                self.add_resource(self.resource0)

        async def stop(self):
            await self.resource_freed(self.resource0)
            self.tg1.cancel_scope.cancel()
            self.tg2.cancel_scope.cancel()
            self.done()
            print("stopped0")

    class Subcomponent1(Component):
        async def start(self):
            self.resource0 = await self.get_resource(Resource0)
            async with create_task_group() as self.tg:
                await self.tg.start(task, "start1", name="start1")
                self.done()
                print("started1")

        async def stop(self):
            self.tg.cancel_scope.cancel()
            self.done()
            print("stopped1")
            self.drop_resource(self.resource0)

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())
            self.add_component(Subcomponent1())

    component0 = Component0()

    async with component0:
        pass

    captured = capsys.readouterr()
    assert captured.out.splitlines() == [
        "prepare0",
        "prepared0",
        "start0",
        "started0",
        "start1",
        "started1",
        "stopped1",
        "stopped0",
    ]


async def test_failing_task(capsys):

    async def failing_task():
        await sleep(0.1)
        raise RuntimeError("start0")

    class Subcomponent0(Component):
        async def start(self):
            async with create_task_group() as self.tg:
                self.tg.start_soon(failing_task)
                self.done()
                print("started0")

    class Subcomponent1(Component):
        pass

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())
            self.add_component(Subcomponent1())

    component0 = Component0()

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass
    
    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "start0"

    captured = capsys.readouterr()
    assert captured.out.splitlines() == [
        "started0",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

from anyio import Event


if TYPE_CHECKING:
    from ._component import Component  # pragma: no cover


class Resource:
    def __init__(self, value: Any, owner: "Component", exclusive: bool):
        self._value = value
        self._owner = owner
        self._exclusive = exclusive
        self._borrowers: set[Component] = set()
        self._dropped = Event()

    def drop(self, borrower: "Component"):
        if borrower in self._borrowers:
            self._borrowers.remove(borrower)
            self._dropped.set()
            self._dropped = Event()

    async def wait_no_borrower(self):
        if not self._borrowers:
            return
        await self._dropped.wait()


class Context:
    def __init__(self):
        self._context = {}
        self._resource_added = Event()

    def add_resource(self, resource: Any, owner: "Component", types: Iterable | Any | None = None, exclusive: bool = False) -> Resource:
        _resource = Resource(resource, owner, exclusive)
        if types is None:
            types = [type(resource)]
        try:
            for resource_type in types:
                break
        except TypeError as e:
            types = [types]
        for resource_type in types:
            if resource_type in self._context:
                raise RuntimeError(f'Resource type "{resource_type}" already exists')
            self._context[resource_type] = _resource
            self._resource_added.set()
            self._resource_added = Event()
        return _resource

    async def get_resource(self, resource_type: Any, borrower: "Component") -> Resource:
        while True:
            if resource_type in self._context:
                resource = self._context[resource_type]
                if resource._exclusive:
                    while True:
                        if not resource._borrowers:
                            resource._borrowers.add(borrower)
                            return resource
                        await resource.wait_no_borrower()
                resource._borrowers.add(borrower)
                return resource
            await self._resource_added.wait()



import sys

import pytest

from fastaio import Component

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_prepare():

    class Component0(Component):
        async def prepare(self):
            # will never prepare because doesn't call self.done()
            pass

    component0 = Component0(name="component0", prepare_timeout=0.1)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(TimeoutError)
    assert str(excinfo.value.exceptions[0]) == "Component timed out while preparing: component0"


async def test_nested_prepare():

    class Subcomponent0(Component):
        async def prepare(self):
            # will never prepare because doesn't call self.done()
            pass

    class Component0(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.subcomponent0 = self.add_component(Subcomponent0(), name="subcomponent0")

    component0 = Component0(name="component0", prepare_timeout=0.1)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(TimeoutError)
    assert str(excinfo.value.exceptions[0]) == "Component timed out while preparing: component0.subcomponent0"


async def test_start():

    class Component0(Component):
        async def start(self):
            # will never start because doesn't call self.done()
            pass

    component0 = Component0(name="component0", start_timeout=0.1)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(TimeoutError)
    assert str(excinfo.value.exceptions[0]) == "Component timed out while starting: component0"


async def test_nested_start():

    class Subcomponent0(Component):
        async def start(self):
            # will never start because doesn't call self.done()
            pass

    class Component0(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.subcomponent0 = self.add_component(Subcomponent0(*args, **kwargs), name="subcomponent0")

    component0 = Component0(name="component0", start_timeout=0.1)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(TimeoutError)
    assert str(excinfo.value.exceptions[0]) == "Component timed out while starting: component0.subcomponent0"


async def test_stop():

    class Component0(Component):
        async def stop(self):
            # will never stop because doesn't call self.done()
            pass

    component0 = Component0(name="component0", stop_timeout=0.1)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(TimeoutError)
    assert str(excinfo.value.exceptions[0]) == "Component timed out while stopping: component0"

async def test_nested_stop():

    class Subcomponent0(Component):
        async def stop(self):
            # will never stop because doesn't call self.done()
            pass

    class Component0(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.subcomponent0 = self.add_component(Subcomponent0(*args, **kwargs), name="subcomponent0")

    component0 = Component0(name="component0", stop_timeout=0.1)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with component0:
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(TimeoutError)
    assert str(excinfo.value.exceptions[0]) == "Component timed out while stopping: component0.subcomponent0"



from __future__ import annotations

import sys

from contextlib import AsyncExitStack
from inspect import isawaitable
from typing import Any, Iterable

import anyio
from anyio import Event, create_task_group, move_on_after
from anyioutils import create_task, wait, FIRST_COMPLETED

from ._context import Context

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

class Component:

    def __init__(
        self,
        name: str | None = None,
        prepare_timeout=1,
        start_timeout=1,
        stop_timeout=1,
    ):
        self._initialized = True
        self._prepare_timeout = prepare_timeout
        self._start_timeout = start_timeout
        self._stop_timeout = stop_timeout
        self._parent = None
        self._context = Context()
        self._prepared = Event()
        self._started = Event()
        self._stopped = Event()
        if name is None:
            self._name = str(self)
        else:
            self._name = name
        self._path = []
        self._components = {}
        self._added_resources = {}
        self._acquired_resources = {}
        self._context_manager_exits = []

    @property
    def path(self) -> str:
        return ".".join(self._path + [self._name])

    def _check_initialized(self):
        try:
            self._initialized
        except AttributeError:
            raise RuntimeError("You must call super().__init__() in the __init__ method of your component")

    def add_component(self, component: "Component", name: str | None = None) -> "Component":
        self._check_initialized()
        if name is None:
            name = component._name
        else:
            component._name = name
        component._path = self._path + [self._name]
        if name in self._components:
            raise RuntimeError(f"Component name already exists: {name}")
        self._components[name] = component
        component._parent = self
        return component

    async def resource_freed(self, resource: Any) -> None:
        await self._added_resources[resource].wait_no_borrower()

    async def all_resources_freed(self) -> None:
        for resource in self._added_resources.values():
            await resource.wait_no_borrower()

    def drop_resource(self, resource: Any) -> None:
        self._acquired_resources[resource].drop(self)

    def add_resource(self, resource: Any, types: Iterable | Any | None = None, exclusive: bool = False) -> None:
        self._added_resources[resource] = self._context.add_resource(resource, self, types, exclusive)
        if self._parent is not None:
            self._added_resources[resource] = self._parent._context.add_resource(resource, self, types, exclusive)

    async def get_resource(self, resource_type: Any) -> Any:
        async with create_task_group() as tg:
            tasks = [create_task(self._context.get_resource(resource_type, self), tg)]
            if self._parent is not None:
                tasks.append(create_task(self._parent._context.get_resource(resource_type, self), tg))
            done, pending = await wait(tasks, tg, return_when=FIRST_COMPLETED)
            for task in done:
                break
            resource = await task.wait()
            tg.cancel_scope.cancel()
        self._acquired_resources[resource._value] = resource
        return resource._value

    async def __aenter__(self):
        self._check_initialized()
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(create_task_group())
            self._exceptions = []
            self._phase = "preparing"
            with move_on_after(self._prepare_timeout) as scope:
                self._task_group.start_soon(self._prepare, name=f"{self.path} prepare")
                await self._all_prepared()
            if scope.cancelled_caught:
                self._get_all_prepare_timeout()
            if not self._exceptions:
                self._phase = "starting"
                with move_on_after(self._start_timeout) as scope:
                    self._task_group.start_soon(self._start, name=f"{self.path} start")
                    await self._all_started()
                if scope.cancelled_caught:
                    self._get_all_start_timeout()
            self._exit_stack = exit_stack.pop_all()
            return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._phase = "stopping"
        with move_on_after(self._stop_timeout) as scope:
            self._task_group.start_soon(self._stop, name=f"{self.path} stop")
            await self._all_stopped()
        if scope.cancelled_caught:
            self._get_all_stop_timeout()
        await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)
        exceptions = []
        for exc in self._exceptions:
            while isinstance(exc, ExceptionGroup):
                exc = exc.exceptions[0]
            exceptions.append(exc)
        if exceptions:
            raise ExceptionGroup("Application failed", exceptions)

    def context_manager(self, resource):
        self._context_manager_exits.append(resource.__exit__)
        return resource.__enter__()

    async def async_context_manager(self, resource):
        self._context_manager_exits.append(resource.__aexit__)
        return await resource.__aenter__()

    def _get_all_prepare_timeout(self):
        for component in self._components.values():
            component._get_all_prepare_timeout()
        if not self._prepared.is_set():
            self._exceptions.append(TimeoutError(f"Component timed out while preparing: {self.path}"))

    def _get_all_start_timeout(self):
        for component in self._components.values():
            component._get_all_start_timeout()
        if not self._started.is_set():
            self._exceptions.append(TimeoutError(f"Component timed out while starting: {self.path}"))

    def _get_all_stop_timeout(self):
        for component in self._components.values():
            component._get_all_stop_timeout()
        if not self._stopped.is_set():
            self._exceptions.append(TimeoutError(f"Component timed out while stopping: {self.path}"))

    async def _all_prepared(self):
        for component in self._components.values():
            await component._all_prepared()
        await self._prepared.wait()

    async def _all_started(self):
        for component in self._components.values():
            await component._all_started()
        await self._started.wait()

    async def _all_stopped(self):
        for component in self._components.values():
            await component._all_stopped()
        await self._stopped.wait()

    async def _prepare(self) -> None:
        try:
            async with create_task_group() as tg:
                for component in self._components.values():
                    component._task_group = tg
                    component._phase = self._phase
                    component._exceptions = self._exceptions
                    tg.start_soon(component._prepare, name=f"{component.path} prepare")
                tg.start_soon(self.prepare, name=f"{self.path} prepare")
        except Exception as exc:
            self._exceptions.append(exc)
            self._prepared.set()

    async def prepare(self) -> None:
        self.done()

    def done(self) -> None:
        if self._phase == "preparing":
            self._prepared.set()
        elif self._phase == "starting":
            self._started.set()
        else:
            for resource in self._acquired_resources.values():
                resource.drop(self)
            for resource in self._added_resources.values():
                if resource._borrowers:
                    raise RuntimeError(f"Resource was not freed: {resource._value}")
            self._stopped.set()

    async def _start(self) -> None:
        try:
            async with create_task_group() as tg:
                for component in self._components.values():
                    component._task_group = tg
                    component._phase = self._phase
                    tg.start_soon(component._start, name=f"{component.path} start")
                tg.start_soon(self.start, name=f"{self.path} start")
        except Exception as exc:
            self._exceptions.append(exc)
            self._started.set()

    async def start(self) -> None:
        self.done()

    async def _stop(self) -> None:
        try:
            async with create_task_group() as tg:
                for component in self._components.values():
                    component._task_group = tg
                    component._phase = self._phase
                    tg.start_soon(component._stop, name=f"{component.path} stop")
                for context_manager_exit in self._context_manager_exits[::-1]:
                    res = context_manager_exit(None, None, None)
                    if isawaitable(res):
                        await res   
                tg.start_soon(self.stop, name=f"{self.path} stop")
        except Exception as exc:
            self._exceptions.append(exc)
            self._stopped.set()

    async def stop(self) -> None:
        self.done()

    async def _main(self): # pragma: no cover
        async with self:
            await Event().wait()

    def run(self):  # pragma: no cover
        try:
            anyio.run(self._main)
        except KeyboardInterrupt:
            pass



import sys

import pytest

from anyio import sleep
from fastaio import Component

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pragma: no cover

pytestmark = pytest.mark.anyio


async def test_resource():

    resource0 = None
    resource1 = None

    class Resource0:
        pass

    class Resource1:
        pass

    class Subcomponent0(Component):
        async def start(self):
            nonlocal resource1
            self.resource0 = Resource0()
            self.add_resource(self.resource0)
            self.resource1 = resource1 = await self.get_resource(Resource1)
            self.done()

        async def stop(self):
            self.drop_resource(self.resource1)
            await self.resource_freed(self.resource0)
            self.done()

    class Subcomponent1(Component):
        async def start(self):
            nonlocal resource0
            self.resource0 = resource0 = await self.get_resource(Resource0)
            self.resource1 = Resource1()
            self.add_resource(self.resource1)
            self.done()

        async def stop(self):
            await self.resource_freed(self.resource1)
            self.drop_resource(self.resource0)
            self.done()

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.subcomponent0 = self.add_component(Subcomponent0())
            self.subcomponent1 = self.add_component(Subcomponent1())

    component0 = Component0()

    async with component0:
        pass

    assert type(component0.subcomponent1.resource0) == Resource0
    assert type(component0.subcomponent0.resource1) == Resource1
    assert component0.subcomponent1.resource0 == resource0
    assert component0.subcomponent0.resource1 == resource1


async def test_resource_with_context_manager(capsys):

    class Resource0:
        async def __aenter__(self):
            print("aenter")
            return self

        async def __aexit__(self, exc_type, exc_value, exc_tb):
            print("aexit")

        def __enter__(self):
            print("enter")
            return self

        def __exit__(self, exc_type, exc_value, exc_tb):
            print("exit")

    class Component0(Component):
        async def start(self):
            print("start")
            resource0 = Resource0()
            await self.async_context_manager(resource0)
            self.context_manager(resource0)
            self.done()

        async def stop(self):
            self.done()
            print("stop")

    async with Component0():
        pass

    assert capsys.readouterr().out.splitlines() == [
        "start",
        "aenter",
        "enter",
        "exit",
        "aexit",
        "stop",
    ]


async def test_add_resource_with_type():

    class Subcomponent0(Component):
        async def start(self):
            self.add_resource(0, types=int)
            self.done()

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())

        async def start(self):
            self.resource = await self.get_resource(int)
            assert self.resource == 0
            self.done()

        async def stop(self):
            self.drop_resource(self.resource)
            self.done()

    async with Component0():
        pass


async def test_add_same_resource_type():

    class Component0(Component):
        async def start(self):
            self.add_resource(0)
            self.add_resource(0)

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0():
            pass

    assert len(excinfo.value.exceptions) == 1
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == """Resource type "<class 'int'>" already exists"""


async def test_add_exclusive_resource(capsys):

    class Subcomponent0(Component):
        async def start(self):
            self.add_resource(0, exclusive=True)
            self.done()

    class Subcomponent1(Component):
        async def start(self):
            resource = await self.get_resource(int)
            print("get")
            await sleep(0.1)
            print("drop")
            self.drop_resource(resource)
            self.done()

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())
            self.add_component(Subcomponent1())
            self.add_component(Subcomponent1())

    async with Component0():
        pass

    assert capsys.readouterr().out.splitlines() == [
        "get",
        "drop",
        "get",
        "drop",
    ]


async def test_resource_not_freed():

    class Subcomponent0(Component):
        async def start(self):
            self.add_resource(0, types=int)
            self.done()

    class Component0(Component):
        def __init__(self, name, stop_timeout):
            super().__init__(name, stop_timeout=stop_timeout)
            self.add_component(Subcomponent0(), name="subcomponent0")

        async def start(self):
            self.resource = await self.get_resource(int)
            assert self.resource == 0
            self.done()

        async def stop(self):
            pass

    with pytest.raises(ExceptionGroup) as excinfo:
        async with Component0(name="component0", stop_timeout=0.1):
            pass

    assert len(excinfo.value.exceptions) == 2
    assert excinfo.group_contains(RuntimeError)
    assert str(excinfo.value.exceptions[0]) == "Resource was not freed: 0"
    assert str(excinfo.value.exceptions[1]) == "Component timed out while stopping: component0"


async def test_all_resources_freed(capsys):

    class Subcomponent0(Component):
        async def start(self):
            self.add_resource(0, types=int)
            self.done()

        async def stop(self):
            await self.all_resources_freed()
            print("all resources freed")
            self.done()

    class Component0(Component):
        def __init__(self):
            super().__init__()
            self.add_component(Subcomponent0())

        async def start(self):
            self.resource = await self.get_resource(int)
            assert self.resource == 0
            self.done()

        async def stop(self):
            self.drop_resource(self.resource)
            print("resource dropped")
            self.done()

    async with Component0():
        pass

    assert capsys.readouterr().out.splitlines() == [
        "resource dropped",
        "all resources freed",
    ]
